"""TimeStamp module."""

from datetime import datetime
from dateutil.parser import parse as dtparse

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class TimeStamp(TimePrimitive):
    """A KML `<TimeStamp>` tag constructor.

    Represents a single moment in time. This is a simple element and contains no
    children. Its value is a `datetime`, specified in XML time. The precision of the
    `TimeStamp` is dictated by the value of `when`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#style
    * http://www.w3.org/TR/xmlschema-2/#isoformats

    Parameters
    ----------
    when : datetime | str

    Attributes
    ----------
    when : datetime

    """

    _kml_tag = "TimeStamp"
    _kml_fields = TimePrimitive._kml_fields + (_FieldDef("when"),)

    def __init__(
        self,
        when: datetime | str,
    ):
        """TimeStamp instance constructor."""
        TimePrimitive.__init__(self)
        self.when: datetime
        if isinstance(when, str):
            self.when = dtparse(when)
        else:
            self.when = when


class GxTimeStamp(TimeStamp):
    """Version of `TimeStamp` under the `gx` namespace.

    Available for use by `AbstractView` subclasses.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gx:timespan-and-gx:timestamp

    """

    _kml_tag = "gx:TimeStamp"



"""AbstractView module."""

from abc import ABC
from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.types import ViewerOption
from pyLiveKML.utils import with_ns
from pyLiveKML.objects.Object import Object, _DependentDef
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class AbstractView(Object, ABC):
    """A KML `<AbstractView>` tag constructor.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#abstractview

    Parameters
    ----------
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        Timestamp or timespan assigned to the object.

    Attributes
    ----------
    viewer_options : list[ViewerOption]
    time_primitive : TimePrimitive | None

    """

    _kml_dependents = Object._kml_dependents + (_DependentDef("time_primitive"),)

    def __init__(
        self,
        viewer_options: Iterable[ViewerOption] | ViewerOption | None = None,
        time_primitive: TimePrimitive | None = None,
    ) -> None:
        """AbstractView instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)
        self._viewer_options = list[ViewerOption]()
        if viewer_options is not None:
            if isinstance(viewer_options, ViewerOption):
                self._viewer_options.append(viewer_options)
            else:
                self._viewer_options.extend(viewer_options)
        self.time_primitive = time_primitive

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element."""
        super().build_kml(root, with_children, with_dependents)
        if self._viewer_options:
            opts = etree.SubElement(root, with_ns("gx:ViewerOptions"))
            for v in self._viewer_options:
                v.build_kml(opts)



"""TimeSpan module."""

from datetime import datetime
from dateutil.parser import parse as dtparser

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class TimeSpan(TimePrimitive):
    """A KML `<TimeSpan>` tag constructor.

    Represents an extent in time bounded by begin and end `datetime`s.

    Notes
    -----
    If `begin` or `end` is `None`, then that end of the period is unbounded.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#style

    Parameters
    ----------
    begin: datetime | str | None, default = None
    end: datetime | str | None, default = None

    Attributes
    ----------
    begin: datetime | None
    end: datetime | None

    """

    _kml_tag = "TimeSpan"
    _kml_fields = TimePrimitive._kml_fields + (
        _FieldDef("begin"),
        _FieldDef("end"),
    )

    def __init__(
        self,
        begin: datetime | str | None = None,
        end: datetime | str | None = None,
    ):
        """TimeSpan instance constructor."""
        TimePrimitive.__init__(self)
        self.begin: datetime | None
        self.end: datetime | None
        if isinstance(begin, str):
            self.begin = dtparser(begin)
        else:
            self.begin = begin
        if isinstance(end, str):
            self.end = dtparser(end)
        else:
            self.end = end


class GxTimeSpan(TimeSpan):
    """Version of `TimeSpan` under the `gx` namespace.

    Available for use by `AbstractView` subclasses.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gx:timespan-and-gx:timestamp

    """

    _kml_tag = "gx:TimeSpan"



"""Region module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import AltitudeModeEnum
from pyLiveKML.objects.Object import _BaseObject, _DependentDef, _FieldDef, Object


class LatLonAltBox(_BaseObject):
    """A KML `<LatLonAltBox>` tag constructor.

    A bounding box that describes an area of interest defined by geographic coordinates
    and altitudes.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-region

    Parameters
    ----------
    north : float
        Specifies the latitude of the north edge of the bounding box, in decimal degrees.
    south : float
        Specifies the latitude of the south edge of the bounding box, in decimal degrees.
    east : float
        Specifies the longitude of the east edge of the bounding box, in decimal degrees.
    west : float
        Specifies the longitude of the west edge of the bounding box, in decimal degrees.
    min_altitude : float, default = 0
        Specified in meters (and is affected by the altitude mode specification).
    max_altitude : float, default = 0
        Specified in meters (and is affected by the altitude mode specification).
    altitude_mode : AltitudeModeEnum, default = AltitudeModeEnum.CLAMP_TO_GROUND

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "LatLonAltBox"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("north"),
        _FieldDef("south"),
        _FieldDef("east"),
        _FieldDef("west"),
        _FieldDef("min_altitude", "minAltitude"),
        _FieldDef("max_altitude", "maxAltitude"),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )

    def __init__(
        self,
        north: float,
        south: float,
        east: float,
        west: float,
        min_altitude: float = 0,
        max_altitude: float = 0,
        altitude_mode: AltitudeModeEnum = AltitudeModeEnum.CLAMP_TO_GROUND,
    ):
        """LatLonAltBox instance constructor."""
        super().__init__()
        self.north = north
        self.south = south
        self.east = east
        self.west = west
        self.min_altitude = min_altitude
        self.max_altitude = max_altitude
        self.altitude_mode = altitude_mode


class LevelOfDetail(_BaseObject):
    """A KML `<Lod>` tag constructor.

    `<Lod>` describes the size of the projected region on the screen that is required in
    order for the region to be considered "active." Also specifies the size of the
    pixel ramp used for fading in (from transparent to opaque) and fading out (from
    opaque to transparent).

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-region
    * https://developers.google.com/kml/documentation/regions
    * https://www.google.com/earth/outreach/tutorials/region.html

    Parameters
    ----------
    min_lod_pixels : float, default = 256
        Defines a square in screen space, with sides of the specified value in pixels.
        For example, 128 defines a square of 128 x 128 pixels. The `region`'s bounding
        box must be larger than this square (and smaller than the `max_lod_pixels`
        square) in order for the `Region` to be active.
    max_lod_pixels : float, default = -1
        Measurement in screen pixels that represents the maximum limit of the visibility
        range for a given Region. A value of -1, the default, indicates "active to
        infinite size."
    min_fade_extent : float, default = 0
        Distance over which the geometry fades, from fully opaque to fully transparent.
        This ramp value, expressed in screen pixels, is applied at the minimum end of the
        LOD (visibility) limits
    max_fade_extent : float, default = 0
        Distance over which the geometry fades, from fully transparent to fully opaque.
        This ramp value, expressed in screen pixels, is applied at the maximum end of the
        LOD (visibility) limits.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Lod"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("min_lod_pixels", "minLodPixels"),
        _FieldDef("max_lod_pixels", "maxLodPixels"),
        _FieldDef("min_fade_extent", "minFadeExtent"),
        _FieldDef("max_fade_extent", "maxFadeExtent"),
    )

    def __init__(
        self,
        min_lod_pixels: float = 256,
        max_lod_pixels: float = -1,
        min_fade_extent: float = 0,
        max_fade_extent: float = 0,
    ):
        """Lod instance constructor."""
        super().__init__()
        self.min_lod_pixels = min_lod_pixels
        self.max_lod_pixels = max_lod_pixels
        self.min_fade_extent = min_fade_extent
        self.max_fade_extent = max_fade_extent


class Region(Object):
    """A KML `<Region>` tag constructor.

    A `Region` contains a bounding box (`LatLonAltBox`) that describes an area of
    interest defined by geographic coordinates and altitudes. In addition, a `Region`
    contains an LOD (level of detail) extent (`LevelOfDetail`) that defines a validity
    range of the associated `Region` in terms of projected screen size. A `Region` is
    said to be "active" when the bounding box is within the user's view and the LOD
    requirements are met. Objects associated with a `Region` are drawn only when the
    `Region` is active. When the `view_refresh_mode` is `ON_REGION`, the `Link` or `Icon`
    is loaded only when the `Region` is active. See the "Topics in KML" page on Regions
    for more details. In a `Container` or `NetworkLink` hierarchy, this calculation uses
    the `Region` that is the closest ancestor in the hierarchy.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference
    * https://developers.google.com/kml/documentation/regions

    Parameters
    ----------
    box : LatLonAltBox
        A bounding box that describes an area of interest defined by geographic
        coordinates and altitudes.
    lod : LevelOfDetail
        Describes the size of the projected region on the screen that is required in
        order for the region to be considered "active."

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Region"
    _kml_depdendents = Object._kml_dependents + (
        _DependentDef("box"),
        _DependentDef("lod"),
    )

    def __init__(
        self,
        box: LatLonAltBox,
        lod: LevelOfDetail,
    ) -> None:
        """Region instance constructor."""
        Object.__init__(self)
        self.box = box
        self.lod = lod



"""Definitions of exceptions and/or errors."""

##########################
#
# pyLiveKML base error class
#
##########################


class KMLError(Exception):
    """Wrapper class for all pyLiveKML errors."""


##########################
#
# Feature errors
#
##########################


class FeatureError(KMLError):
    """Wrapper class for errors with :class:`pyLiveKML.objects.Feature` classes and subclasses."""

    pass


class FeatureInaccessibleError(FeatureError):
    """Indicates that a :class:`pyLiveKML.objects.Feature` was not able to be updated.

    This is typically because the `Feature` instance is visible in the UI. Only `Feature`
    instances that are not visible may be updated.
    """

    pass


##########################
#
# LinearRing errors
#
##########################


class LinearRingError(KMLError):
    """Wrapper class for errors with :class:`pyLiveKML.objects.LinearRing`."""

    pass


class LinearRingCoordsError(LinearRingError):
    """Indicates that the :attr:`pyLiveKML.objects.LinearRing.coordinates` parameter was invalid.

    This is typically because less than the minimum of 3 coordinates were supplied.
    """


##########################
#
# LineString errors
#
##########################


class LineStringError(KMLError):
    """Wrapper class for errors with :class:`pyLiveKML.objects.LineString`."""

    pass


class LineStringCoordsError(LineStringError):
    """Indicates that the :attr:`pyLiveKML.objects.LineString.coordinates` parameter was invalid.

    This is typically because less than the minimum of 2 coordinates were supplied.
    """


##########################
#
# NetworkLinkControl errors
#
##########################


class NetworkLinkControlError(KMLError):
    """Wrapper class for errors with :class:`pyLiveKML.objects.NetworkLinkControl` classes and subclasses."""

    pass


class NetworkLinkControlUpdateLimited(NetworkLinkControlError):
    """Indicates that the maximum permitted number of entries has been added to a :class:`pyLiveKML.objects.Update` instance.

    Primarily intended to be used to break out of a recursive loop in
    :function:`pyLiveKML.objects.NetworkLinkControl.build_kml()` when the maximum update
    size has been reached.
    """

    pass


##########################
#
# Track errors
#
##########################


class TrackError(FeatureError):
    """Wrapper class for errors with :class:`pyLiveKML.objects.Track` classes and subclasses."""

    pass


class TrackElementsMismatch(TrackError):
    """Indicates that there was a problem with the `elements` data passed to a :class:`pyLiveKML.objects.Track`.

    This typically occurs when the `elements` data has mismatched lengths for the
    elements of its `extended_data` fields.
    """

    pass



"""ViewerOption module."""

from lxml import etree  # type: ignore

from pyLiveKML.types.types import ViewerOptionEnum
from pyLiveKML.utils import with_ns
from pyLiveKML.objects.Object import _BaseObject


class ViewerOption(_BaseObject):
    """Enables or disables special viewing modes.

    Used only in conjunction with subclasses of :class:`pyLiveKML.objects.AbstractView`.

    Notes
    -----
    * Applies only to Google Earth 6.0 and later.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-abstractview.

    Parameters
    ----------
    option : ViewerOptionEnum
        The view option to be configured.
    enabled : bool
        Whether the view option is to be enabled or disabled.

    """

    _kml_tag = "gx:option"

    def __init__(self, option: ViewerOptionEnum, enabled: bool):
        """GxViewerOption instance constructor."""
        super().__init__()
        self.option = option
        self.enabled = enabled

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element."""
        super().build_kml(root, with_children, with_dependents)
        etree.SubElement(
            root,
            with_ns(self._kml_tag),
            attrib={"name": self.option.value, "enabled": str(int(self.enabled))},
        )



"""LookAt module."""

from typing import Sequence

from lxml import etree  # type: ignore

from pyLiveKML.types import AltitudeModeEnum, ViewerOption
from pyLiveKML.objects.Object import (
    _FieldDef,
    Angle180,
    Angle360,
    Angle90,
    AnglePos90,
    NoParse,
)
from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class LookAt(AbstractView):
    """A KML `<LookAt>` tag constructor.

    Defines a virtual camera that is associated with any element derived from `Feature`.
    The `LookAt` element positions the "camera" in relation to the object that is being
    viewed. In Google Earth, the view "flies to" this `LookAt` viewpoint when the user
    double-clicks an item in the "Places" panel or double-clicks an icon in the 3D
    viewer.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#lookat

    Parameters
    ----------
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        Timestamp or timespan assigned to the object.
    lla : tuple[float, float, float], default = (0, 0, 0)
        The longitude, latitude and altitude (in that order, in decimal degrees) of the
        camera position.
    angles : tuple[float, float], default (0, 0)
        The heading, tilt and roll (in that order, in decimal degrees) of the `LookAt`
        facing.
    range : float
        Distance in meters from the point specified by `longitude`, `latitude` and
        `altitude` to the `LookAt` position.
    altitude_mode : AltitudeModeEnum | None, default = None

    Attributes
    ----------
    viewer_options : list[ViewerOption]
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None
        Timestamp or timespan assigned to the object.
    longitude : float
        The longitude of the `LookAt` position, in decimal degrees.
    latitude : float
        The latitude of the `LookAt` position, in decimal degrees.
    altitude : float
        The altitude of the `LookAt` position, in metres, with respect to `altitude_mode`.
    heading : float
        The heading of the `LookAt` facing, in decimal degrees.
    tilt : float
        The tilt of the `LookAt` facing, in decimal degrees.
    range : float
        Distance in meters from the point specified by `longitude`, `latitude` and
        `altitude` to the `LookAt` position.
    altitude_mode : AltitudeModeEnum | None

    """

    _kml_tag = "LookAt"
    _kml_fields = AbstractView._kml_fields + (
        _FieldDef("longitude", parser=Angle180),
        _FieldDef("latitude", parser=Angle90),
        _FieldDef("altitude"),
        _FieldDef("heading", parser=Angle360),
        _FieldDef("tilt", parser=AnglePos90),
        _FieldDef("range", parser=NoParse),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )

    def __init__(
        self,
        viewer_options: Sequence[ViewerOption] | ViewerOption | None = None,
        time_primitive: TimePrimitive | None = None,
        lla: tuple[float, float, float] = (0, 0, 0),
        angles: tuple[float, float] = (0, 0),
        range: float = 0,
        altitude_mode: AltitudeModeEnum | None = None,
    ):
        """LookAt instance constructor."""
        AbstractView.__init__(self, viewer_options, time_primitive)
        self.longitude, self.latitude, self.altitude = lla
        self.heading, self.tilt = angles
        self.range = range
        self.altitude_mode = (
            AltitudeModeEnum.CLAMP_TO_GROUND if altitude_mode is None else altitude_mode
        )



"""IconStyle module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import ColorModeEnum, HotSpot, GeoColor
from pyLiveKML.objects.Object import _BaseObject, _FieldDef, NoDump, _DependentDef
from pyLiveKML.objects.ColorStyle import ColorStyle


class _IconStyle_Icon(_BaseObject):
    """A minimalist Icon class, used only within `IconStyle`."""

    _kml_tag = "Icon"
    _kml_fields = _BaseObject._kml_fields + (_FieldDef("href"),)

    def __init__(self, href: str):
        """_IconStyle_Icon instance constructor."""
        super().__init__()
        self.href = href


class IconStyle(ColorStyle):
    """A KML `<IconStyle>` tag constructor.

    Specifies how icons for `Point` placemarks are drawn, both in the "Places" panel and
    in the 3D viewer of Google Earth. The `icon` attribute specifies the icon image. The
    `scale` attribute specifies the x, y scaling of the icon. The color specified in the
    `color` attribute of `IconStyle` is blended with the color of the `Icon`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#iconstyle

    Parameters
    ----------
    icon: str
        An HTTP address or a local file specification used to load an icon.
    scale: float, default = 1.0
        Resizes the icon.
    heading: float | None, default = None
        Direction in decimal degrees. If not specified, defaults to 0. Values range from
        0 to 360 degrees.
    color: GeoColor | int | None, default = None
    color_mode: ColorModeEnum | None, default = None
    hot_spot: HotSpot | None, default = None
        Specifies the position within the icon that is "anchored" to the `Point` to which
        the style is applied.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "IconStyle"
    _kml_fields = ColorStyle._kml_fields + (
        _FieldDef("icon", dumper=NoDump),
        _FieldDef("scale"),
        _FieldDef("heading"),
    )
    _kml_dependents = ColorStyle._kml_dependents + (
        _DependentDef("icon"),
        _DependentDef("hot_spot"),
    )

    def __init__(
        self,
        icon: str,
        scale: float = 1.0,
        heading: float | None = None,
        color: GeoColor | int | None = None,
        color_mode: ColorModeEnum | None = None,
        hot_spot: HotSpot | None = None,
    ):
        """IconStyle instance constructor."""
        ColorStyle.__init__(self, color=color, color_mode=color_mode)
        self.scale = scale
        self.heading = heading
        self.icon = _IconStyle_Icon(icon)
        self.hot_spot = hot_spot



"""ColorStyle module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import ColorModeEnum, GeoColor
from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.ColorStyle import ColorStyle


class LabelStyle(ColorStyle):
    """A KML `<LabelStyle>` tag constructor.

    Specifies how the `name` of a `Feature` is drawn in the 3D viewer.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#labelstyle

    Parameters
    ----------
    scale: float, default = 1.0
        Resizes the label.
    color: GeoColor | int | None, default = None
    color_mode: ColorModeEnum | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "LabelStyle"
    _kml_fields = ColorStyle._kml_fields + (_FieldDef("scale"),)

    def __init__(
        self,
        scale: float | None = None,
        color: GeoColor | int | None = None,
        color_mode: ColorModeEnum | None = None,
    ):
        """ColorStyle instance constructor."""
        super().__init__(color=color, color_mode=color_mode)
        self.scale = scale



"""Model module."""

from typing import Iterator, Iterable

from lxml import etree  # type: ignore

from pyLiveKML.types import AltitudeModeEnum
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _FieldDef,
    _ListObject,
    Angle90,
    Angle180,
    AnglePos180,
    Angle360,
    Object,
)
from pyLiveKML.objects.Link import Link
from pyLiveKML.objects.Geometry import Geometry


class Location(_BaseObject):
    """A KML `<Location>` tag constructor.

    Specific to :class:`pyLiveKML.objects.Model.Model`.

    Specifies the exact coordinates of the `Model`'s origin in latitude, longitude, and
    altitude. Latitude and longitude measurements are standard lat-lon projection with
    WGS84 datum. Altitude is distance above the earth's surface, in meters, and is
    interpreted according to the `Model`'s `altitude_mode`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-model

    Parameters
    ----------
    longitude : float, default = 0
    latitude : float, default = 0
    altitude : float, default = 0

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Location"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("longitude", parser=Angle180),
        _FieldDef("latitude", parser=Angle90),
        _FieldDef("altitude"),
    )

    def __init__(self, longitude: float = 0, latitude: float = 0, altitude: float = 0):
        """Location instance constructor."""
        super().__init__()
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude


class Orientation(_BaseObject):
    """A KML `<Orientation>` tag constructor.

    Specific to :class:`pyLiveKML.objects.Model.Model`.

    Describes rotation of a 3D model's coordinate system to position the object in Google
    Earth.

    Rotations are applied to a Model in the following order:
    * roll
    * tilt
    * heading

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-model

    Parameters
    ----------
    heading : float, default = 0
    tilt : float, default = 0
    roll : float, default = 0

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Orientation"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("heading", parser=Angle360),
        _FieldDef("tilt", parser=AnglePos180),
        _FieldDef("roll", parser=Angle180),
    )

    def __init__(self, heading: float = 0, tilt: float = 0, roll: float = 0):
        """Location instance constructor."""
        super().__init__()
        self.heading = heading
        self.tilt = tilt
        self.roll = roll


class Scales(_BaseObject):
    """A KML `<Scale>` tag constructor.

    Specific to :class:`pyLiveKML.objects.Model.Model`.

    Scales a model along the x, y, and z axes in the model's coordinate space.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-model

    Parameters
    ----------
    x : float, default = 0
    y : float, default = 0
    z : float, default = 0

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Scale"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("x"),
        _FieldDef("y"),
        _FieldDef("z"),
    )

    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        """Scale instance constructor."""
        super().__init__()
        self.x = x
        self.y = y
        self.z = z


class Alias(_BaseObject):
    """A KML `<Alias>` tag constructor.

    Specific to :class:`pyLiveKML.objects.Model.Model`.

    A mapping for the texture file paths from the original COLLADA file to the KML file
    that contains the owning `Model`. Allows texture files to be moved and/or renamed
    without having to update the original COLLADA file that references those textures.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-model

    Parameters
    ----------
    target_href : str
        Specifies the texture file to be fetched by Google Earth. This reference can be a
        relative reference to an image file within the .kmz archive, or it can be an
        absolute reference to the file (for example, a URL).
    source_href : str
        Is the path specified for the texture file in the Collada .dae file.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Alias"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("target_href", "targetHref"),
        _FieldDef("source_href", "sourceHref"),
    )

    def __init__(self, target_href: str, source_href: str):
        """Alias instance constructor."""
        super().__init__()
        self.target_href = target_href
        self.source_href = source_href


class ResourceMap(_ListObject[Alias], _BaseObject):
    """A KML `<ResourceMap>` tag constructor.

    Specific to :class:`pyLiveKML.objects.Model.Model`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-model

    Parameters
    ----------
    resources : Alias | Sequence[Alias] | None, default = None
        The `Alias` objects to be contained in the `ResourceMap`

    Attributes
    ----------
    Nil

    """

    _kml_tag = "ResourceMap"
    _kml_children = _BaseObject._kml_children + (_ChildDef("resources"),)

    def __init__(self, resources: Alias | Iterable[Alias] | None = None) -> None:
        """ResourceMap instance constructor."""
        super().__init__()
        self.resources = resources

    @property
    def resources(self) -> Iterator[Alias]:
        """Retrieve a generator over the `Alias`es in this `ResourceMap`.

        If the property setter is called, replaces the current list of contained
        `Alias`es with those provided.

        Parameters
        ----------
        value : Alias | Iterable[Alias] | None
            The new `Alias` elements for the `ResourceMap`.

        :returns: A generator over the `Alias`es in the `ResourceMap`.
        :rtype: Iterator[Alias]

        """
        yield from self

    @resources.setter
    def resources(self, value: Alias | Iterable[Alias] | None) -> None:
        if value is not None:
            if isinstance(value, Alias):
                self.append(value)
            else:
                self.extend(value)


class Model(Geometry):
    """A KML `<Model>` tag constructor.

    A 3D object described in a COLLADA file (referenced in the `link` attribute). COLLADA
    files have a .dae file extension. Models are created in their own coordinate space
    and then located, positioned, and scaled in Google Earth. See the "Topics in KML" page
    on Models (link in References, below) for more detail.

    Google Earth supports the COLLADA common profile, with the following exceptions:
    * Google Earth supports only triangles and lines as primitive types. The maximum
    number of triangles allowed is 21845.
    * Google Earth does not support animation or skinning.
    * Google Earth does not support external geometry references.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#model
    * https://developers.google.com/kml/documentation/models

    Parameters
    ----------
    link : Link
        Specifies the file to load and optional refresh parameters.
    altitude_mode : AltitudeModeEnum | None, default = None
    location : tuple[float, float, float], default = (0, 0, 0)
        The location of the model, as a tuple of (longitude, latitude, altitude).
    orientation : tuple[float, float, float], default = (0, 0, 0)
        The orientation of the model, as a tuple of (heading, tilt, roll).
    scales : tuple[float, float, float], default = (0, 0, 0)
        The scaling of the model, as a tuple of (x, y, z).
    resources : Sequence[Alias] | Alias | None, default = None
        Mappings for the texture file paths from the original COLLADA file to the KML
        file that contains the `Model`.

    Attributes
    ----------
    link : Link
        Specifies the file to load and optional refresh parameters.
    altitude_mode : AltitudeModeEnum | None, default = None
    location : Location
        The location of the model.
    orientation : Orientation
        The orientation of the model.
    scales : Scales
        The scaling of the model.
    resources : ResourceMap
        Mappings for the texture file paths from the original COLLADA file to the KML
        file that contains the `Model`.

    """

    _kml_tag = "Model"
    _kml_fields = Object._kml_fields + (_FieldDef("altitude_mode", "gx:altitudeMode"),)
    _kml_children = Object._kml_children + (
        _ChildDef("link"),
        _ChildDef("location"),
        _ChildDef("orientation"),
        _ChildDef("scale"),
        _ChildDef("resources"),
    )

    def __init__(
        self,
        link: Link,
        altitude_mode: AltitudeModeEnum | None = None,
        location: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float] = (0, 0, 0),
        scales: tuple[float, float, float] = (0, 0, 0),
        resources: Alias | Iterable[Alias] | None = None,
    ) -> None:
        """Model instance constructor."""
        Object.__init__(self)
        self.link = link
        altitude_mode = (
            AltitudeModeEnum.CLAMP_TO_GROUND if altitude_mode is None else altitude_mode
        )
        self.altitude_mode = altitude_mode
        self.location = Location(*location)
        self.orientation = Orientation(*orientation)
        self.scales = Scales(*scales)
        self.resources = ResourceMap(resources)



"""Tour module."""

from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import (
    _BaseObject,
    _DependentDef,
    _FieldDef,
    _ListObject,
    Object,
)
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class Playlist(_ListObject[TourPrimitive], _BaseObject):
    """A KML `<Playlist>` tag constructor.

    Contains any number of `TourPrimitive` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#contains_6

    Parameters
    ----------
    items : TourPrimitive | Iterable[TourPrimitive] | None, default = None

    Attributes
    ----------
    Nil

    """

    _kml_tag = "gx:Playlist"

    def __init__(
        self, items: TourPrimitive | Iterable[TourPrimitive] | None = None
    ) -> None:
        """Playlist instance constructor."""
        _BaseObject.__init__(self)
        _ListObject[TourPrimitive].__init__(self)
        if items is not None:
            if isinstance(items, TourPrimitive):
                self.append(items)
            else:
                self.extend(items)


class Tour(Object):
    """A KML `<gx:Tour>` tag constructor.

    A `Tour` contains a single `PlayList`, which in turn contains an ordered list of
    `TourPrimitive` elements that define a tour in any KML browser.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxtour
    * https://developers.google.com/kml/documentation/touring

    Parameters
    ----------
    name: str | None, default = None
    description: str | None, default = None
    playlist: TourPrimitive | list[TourPrimitive] | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:Tour"
    _kml_fields = Object._kml_fields + (
        _FieldDef("name"),
        _FieldDef("description"),
    )
    _kml_dependents = Object._kml_dependents + (_DependentDef("playlist"),)

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        playlist: TourPrimitive | list[TourPrimitive] | None = None,
    ) -> None:
        """Tour instance constructor."""
        Object.__init__(self)
        self.name = name
        self.description = description
        self.playlist = Playlist(playlist)



"""Update module."""

from abc import ABC
from enum import Enum
from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _BaseObject, _FieldDef, ObjectChild


class UpdateType(Enum):
    """Enumeration of the child tags that may be contained in an `<Update>` tag."""

    CREATE = "Create"
    CHANGE = "Change"
    DELETE = "Delete"


class UpdateSequent(ObjectChild):
    """An element in an Update sequence.

    Parameters
    ----------
    tag : UpdateType
        The `<Update>` sub-tag for the element.
    parent : _BaseObject
        The parent object for any `<Create>` tags. Required for the `id` field.
    child : _BaseObject
        The target object that will be created, changed or deleted.

    Attributes
    ----------
    Same as parameters.

    """

    def __init__(
        self, tag: UpdateType, parent: _BaseObject, child: _BaseObject
    ) -> None:
        """UpdateSequent instance constructor."""
        super().__init__(parent, child)
        self.tag = tag


class _UpdateSequence(_BaseObject, list[UpdateSequent]):
    """A sequence of create, change and delete operations to be executed.

    This sequence will be executed in the order in which it is stored. The primary
    purpose, compared to the :class:`pyLiveKML.objects.Update._UpdateList` subclasses,
    below, is to allow a completely arbitrary sub-tag order for `<Update>` tags. This is
    relevant to :class:`pyLiveKML.objects.AnimatedUpdate`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#update
    * https://developers.google.com/kml/documentation/kmlreference#example-of-create
    * https://developers.google.com/kml/documentation/kmlreference#example-of-change
    * https://developers.google.com/kml/documentation/kmlreference#example-of-delete

    Parameters
    ----------
    items : UpdateSequent | Iterable[UpdateSequent] | None, default = None
        The initial sequence of `UpdateSequent` items.

    """

    _suppress_id = True

    def __init__(
        self,
        items: UpdateSequent | Iterable[UpdateSequent] | None = None,
    ):
        _BaseObject.__init__(self)
        list[UpdateSequent].__init__(self)
        if items is not None:
            if isinstance(items, UpdateSequent):
                self.append(items)
            else:
                self.extend(items)

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element.

        Overridden from `_ObjectBase` to ensure specific behaviour.

        Notes
        -----
        * Unlike the other update list types, **does not** clear the list after
        publishing it.

        """
        for s in self:
            if s.tag == UpdateType.CREATE:
                if not s.parent._suppress_id:
                    create_tag = etree.SubElement(root, s.tag.value)
                    s.child.create_kml(create_tag, s.parent)
            elif s.tag == UpdateType.CHANGE:
                if not s.child._suppress_id:
                    change_tag = etree.SubElement(root, s.tag.value)
                    s.child.change_kml(change_tag)
            elif s.tag == UpdateType.DELETE:
                if not s.child._suppress_id:
                    delete_tag = etree.SubElement(root, s.tag.value)
                    s.child.delete_kml(delete_tag)


class _UpdateList(_BaseObject, list[ObjectChild], ABC):
    """Abstract base for `_CreateList`, `_ChangeList` and `_DeleteList`."""

    def __init__(
        self,
        items: ObjectChild | Iterable[ObjectChild] | None = None,
    ) -> None:
        _BaseObject.__init__(self)
        list[ObjectChild].__init__(self)
        ABC.__init__(self)
        if items is not None:
            if isinstance(items, ObjectChild):
                self.append(items)
            else:
                self.extend(items)


class _CreateList(_UpdateList):
    """A KML `<Create>` tag constructor.

    Always a child of an `Update` instance.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#example-of-create

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    _kml_tag = "Create"

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element.

        Overridden from `_ObjectBase` to ensure specific behaviour, including clearing
        the list after publishing the KML.
        """
        for c in (c for c in self if not c.parent._suppress_id):
            c.child.create_kml(root, c.parent)
        self.clear()


class _ChangeList(_UpdateList):
    """A KML `<Change>` tag constructor.

    Always a child of an `Update` instance.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#example-of-change

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    _kml_tag = "Change"

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element.

        Overridden from `_ObjectBase` to ensure specific behaviour, including clearing
        the list after publishing the KML.
        """
        for c in (c for c in self if not c.child._suppress_id):
            c.child.change_kml(root)
        self.clear()


class _DeleteList(_UpdateList):
    """A KML `<Delete>` tag constructor.

    Always a child of an `Update` instance.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#example-of-delete

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    _kml_tag = "Delete"

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element.

        Overridden from `_ObjectBase` to ensure specific behaviour, including clearing
        the list after publishing the KML.
        """
        for c in (c for c in self if not c.child._suppress_id):
            c.child.delete_kml(root)
        self.clear()


class Update(_BaseObject):
    """A KML `<Update>` tag constructor.

    Specifies additions, changes, or deletions to KML data that has already been loaded
    using the specified URL. `target_href` specifies the .kml or .kmz file whose data
    (within Google Earth) is to be modified. `Update` is always* contained in a
    `NetworkLinkControl`. Furthermore, the file containing the `NetworkLinkControl` must
    have been loaded by a `NetworkLink`.

    Notes
    -----
    * `AnimatedUpdate` may also contain an `Update` tag.
    * The `creates`, `changes` and `deletes` attributes each construct and populate a
    single child tag of the appropriate type, **in that order**. That is, **all** creates
    are emitted, then **all** changes, then **all** deletes. This is optimal for
    pyLiveKML synchronization purposes, but can be problematic for e.g. `AnimatedUpdate`.
    The `sequence` attribute allows an arbitrary ordering of child tags, suitable for
    `AnimatedUpdate`, or any other situation in which a specific ordering of child tags
    is required.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#syntax_50

    Parameters
    ----------
    target_href : str
    creates : ObjectChild | Iterable[ObjectChild] | None, default = None
        A collection of ObjectChild instances to be created.
    changes : ObjectChild | Iterable[ObjectChild] | None, default = None
        A collection of ObjectChild instances to be changed.
    deletes : ObjectChild | Iterable[ObjectChild] | None, default = None
        A collection of ObjectChild instances to be deleted.
    sequence : UpdateSequent | Iterable[UpdateSequent] | None, default = None
        An ordered list of update operations.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Update"
    _kml_fields = _BaseObject._kml_fields + (_FieldDef("target_href", "targetHref"),)

    def __init__(
        self,
        target_href: str,
        creates: ObjectChild | Iterable[ObjectChild] | None = None,
        changes: ObjectChild | Iterable[ObjectChild] | None = None,
        deletes: ObjectChild | Iterable[ObjectChild] | None = None,
        sequence: UpdateSequent | Iterable[UpdateSequent] | None = None,
    ):
        """Update instance constructor."""
        super().__init__()
        self.target_href = target_href
        self.creates = _CreateList(creates)
        self.changes = _ChangeList(changes)
        self.deletes = _DeleteList(deletes)
        self.sequence = _UpdateSequence(sequence)

    def clear(self) -> None:
        """Clear the instance.

        Discards the current contents of the `creates`, `changes`, `deletes` and
        `sequence` lists.
        """
        self.creates.clear()
        self.changes.clear()
        self.deletes.clear()
        self.sequence.clear()

    def __len__(self) -> int:
        """Take the current length of the instance.

        The returned value is the sum of the lengths of the `creates`, `changes` and
        `deletes` lists; it **does not** include the length of the `sequence` list.
        """
        return len(self.creates) + len(self.changes) + len(self.deletes)

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element.

        Overriden from `_BaseObject` to enforce some specific publishing behaviour.
        """
        super().build_kml(root, with_children, with_dependents)
        if self.creates:
            root.append(self.creates.construct_kml(with_children, with_dependents))
        if self.changes:
            root.append(self.changes.construct_kml(with_children, with_dependents))
        if self.deletes:
            root.append(self.deletes.construct_kml(with_children, with_dependents))
        if self.sequence:
            self.sequence.build_kml(root, with_children, with_dependents)



"""Geometry module."""

from abc import ABC

from pyLiveKML.objects.Object import Object


class Geometry(Object, ABC):
    """A KML `<Geometry>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file. It provides a
    placeholder object for all derived `<Geometry>` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#geometry.

    Parameters
    ----------
    Nil

    Attributes
    ----------
    Nil

    """

    def __init__(self) -> None:
        """Geometry instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)



"""Track module."""

from datetime import datetime
from typing import Iterator, Iterable, Any

from dateutil.parser import parse as dtparse
from lxml import etree  # type: ignore

from pyLiveKML.types import AltitudeModeEnum
from pyLiveKML.objects.Object import (
    _ChildDef,
    _FieldDef,
    Angle180,
    Angle360,
    Angle90,
    AnglePos180,
)
from pyLiveKML.errors import TrackElementsMismatch
from pyLiveKML.utils import with_ns
from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Model import Model


class TrackCoord:
    """Coordinates class, specific to the <gx:Track> KML tag."""

    def __init__(
        self,
        longitude: float = 0,
        latitude: float = 0,
        altitude: float = 0,
    ):
        """TrackCoord instance constructor."""
        super().__init__()
        self.longitude = Angle180.parse(longitude)
        self.latitude = Angle90.parse(latitude)
        self.altitude = altitude

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.longitude} {self.latitude} {self.altitude}"


class TrackAngles:
    """Angles (heading/tilt/roll) class, specific to the <gx:Track> KML tag."""

    def __init__(
        self,
        heading: float = 0,
        tilt: float = 0,
        roll: float = 0,
    ) -> None:
        """TrackAngles instance constructor."""
        super().__init__()
        self.heading = Angle360.parse(heading)
        self.tilt = AnglePos180.parse(tilt)
        self.roll = Angle180.parse(roll)

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.heading} {self.tilt} {self.roll}"


class TrackElement:
    """Track element class, specific to the <gx:Track> KML tag.

    Used to collect the various components of each track point that are then distributed
    across the <gx:Track> tag.
    """

    def __init__(
        self,
        when: datetime | str,
        coords: TrackCoord | tuple[float, ...] | None,
        angles: TrackAngles | tuple[float, ...] | None,
        extended_data: dict[str, dict[str, Any]] | None = None,
    ) -> None:
        """TrackElement instance constructor."""
        super().__init__()
        if isinstance(when, datetime):
            self.when = when
        else:
            self.when = dtparse(when)
        self.coords: TrackCoord | None = None
        if coords is None or isinstance(coords, TrackCoord):
            self.coords = coords
        else:
            self.coords = TrackCoord(*coords)
        self.angles: TrackAngles | None = None
        if angles is None or isinstance(angles, TrackAngles):
            self.angles = angles
        else:
            self.angles = TrackAngles(*angles)
        self.extended_data = extended_data


class Track(Geometry):
    """A KML 'gx:Track', per https://developers.google.com/kml/documentation/kmlreference#gxtrack."""

    _kml_tag = "gx:Track"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )
    _kml_children = Geometry._kml_children + (_ChildDef("model"),)

    def __init__(
        self,
        altitude_mode: AltitudeModeEnum | None = None,
        elements: TrackElement | Iterable[TrackElement] | None = None,
        model: Model | None = None,
    ) -> None:
        """Track instance constructor."""
        Geometry.__init__(self)
        self.altitude_mode = altitude_mode
        self.model = model
        self._schemas = dict[str, set[str]]()
        self._elements = list[TrackElement]()
        self.elements = elements

    @property
    def elements(self) -> Iterator[TrackElement]:
        """Retrieve a generator over the set of enclosed `TrackElements`."""
        yield from self._elements

    @elements.setter
    def elements(self, value: TrackElement | Iterable[TrackElement] | None) -> None:
        self._elements.clear()
        self._schemas.clear()
        if value is None:
            return
        if isinstance(value, TrackElement):
            self._elements.append(value)
        else:
            xdlens = set[int]()
            for e in value:
                if e.extended_data is None:
                    xdlens.add(0)
                else:
                    for fname in e.extended_data.values():
                        for fval in fname.values():
                            xdlens.add(0 if fval is None else len(fval))
            if len(xdlens) != 1:
                raise TrackElementsMismatch()
            self._elements.extend(value)
            for xd in (
                e.extended_data for e in self._elements if e.extended_data is not None
            ):
                for schema, fields in xd.items():
                    if schema not in self._schemas:
                        self._schemas[schema] = set[str]()
                    self._schemas[schema].update(fields.keys())

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Construct the KML content and append it to the provided etree.Element."""
        super().build_kml(root, with_children)
        value: str
        for e in self.elements:
            etree.SubElement(root, with_ns("when")).text = e.when.isoformat()
        for e in self.elements:
            value = "" if e.coords is None else str(e.coords)
            etree.SubElement(root, with_ns("gx:coord")).text = value
        for e in self.elements:
            value = "" if e.angles is None else str(e.angles)
            etree.SubElement(root, with_ns("gx:angles")).text = value
        if self._schemas:
            e_xd = etree.SubElement(root, "ExtendedData")
            for sch, fields in self._schemas.items():
                e_sch = etree.SubElement(e_xd, "SchemaData", attrib={"schemaUrl": sch})
                for f in sorted(fields):
                    e_field = etree.SubElement(
                        e_sch,
                        with_ns("gx:SimpleArrayData"),
                        attrib={with_ns("kml:name"): f},
                    )
                    for e in self.elements:
                        value = ""
                        if (
                            e.extended_data
                            and sch in e.extended_data
                            and f in e.extended_data[sch]
                            and e.extended_data[sch][f] is not None
                        ):
                            value = str(e.extended_data[sch][f])
                        etree.SubElement(e_field, with_ns("gx:value")).text = value



"""LineStyle module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef, ColorParse
from pyLiveKML.types.GeoColor import GeoColor
from pyLiveKML.objects.ColorStyle import ColorStyle


class LineStyle(ColorStyle):
    """A KML `<LineStyle>` tag constructor.

    Specifies the drawing style (color, color mode, and line width) for all line
    geometry. Line geometry includes the outlines of outlined `Polygon`s and the extruded
    "tether" of `Placemark` icons (if extrusion is enabled).

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#linestyle

    Parameters
    ----------
    width: float | None, default = None
        Width of the line, in pixels.
    color: GeoColor | int | None, default = None
    outer_color: GeoColor | int | None, default = None
        Color of the portion of the line defined by `outer_width`. Note that the
        `outer_color` and `outer_width` attributes are ignored when a `LineStyle` is
        applied to a `Polygon` or `LinearRing`.
    outer_width: float | None, default = None
        A value between 0.0 and 1.0 that specifies the proportion of the line that uses
        the `outer_color`. Only applies to lines setting width with `physical_width`; it
        does not apply to lines using `width`. See also `draw_order` in `LineString`. A
        draw order value may be necessary if dual-colored lines are crossing each other,
        for example, for showing freeway interchanges.
    physical_width: float | None, default = None
        Physical width of the line, in meters.
    label_visibility: bool | None, default = None
        Whether or not to display a text label on a `LineString`. A `LineString`'s label
        is contained in the `name` attribute that is a sibling of `LineString`, i.e. is
        contained within the same `Placemark`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "LineStyle"
    _kml_fields = ColorStyle._kml_fields + (
        _FieldDef("width"),
        _FieldDef("outer_color", "gx:outerColor", ColorParse),
        _FieldDef("outer_width", "gx:outerWidth"),
        _FieldDef("physical_width", "gx:physicalWidth"),
        _FieldDef("label_visibility", "gx:labelVisibility"),
    )

    def __init__(
        self,
        width: float | None = None,
        color: GeoColor | int | None = None,
        outer_color: GeoColor | int | None = None,
        outer_width: float | None = None,
        physical_width: float | None = None,
        label_visibility: bool | None = None,
    ):
        """LineStyle instance constructor."""
        ColorStyle.__init__(self, color)
        self.width = width
        self.outer_color = outer_color
        self.outer_width = outer_width
        self.physical_width = physical_width
        self.label_visibility = label_visibility



"""TourControl module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import PlayModeEnum
from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class TourControl(TourPrimitive):
    """A KML `<gx:TourControl>` tag constructor.

    Allows a `Tour` to be paused, until the user takes action to continue the `Tour`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#extended-by_8

    Parameters
    ----------
    play_mode : PlayModeEnum, default = PlayModeEnum.PAUSE

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:TourControl"
    _kml_fields = TourPrimitive._kml_fields + (_FieldDef("play_mode", "gx:playMode"),)

    def __init__(
        self,
        play_mode: PlayModeEnum = PlayModeEnum.PAUSE,
    ) -> None:
        """GxTourControl instance constructor."""
        TourPrimitive.__init__(self)
        self.play_mode = play_mode



"""Camera module."""

from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.types import AltitudeModeEnum, ViewerOption
from pyLiveKML.objects.Object import (
    _FieldDef,
    Angle90,
    Angle180,
    AnglePos180,
    Angle360,
)
from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class Camera(AbstractView):
    """A KML `<Camera>` tag constructor.

    Defines the virtual camera that views the scene. This element defines the position of
    the camera relative to the Earth's surface as well as the viewing direction of the
    camera. The camera position is defined by `<longitude>`, `<latitude>`, `<altitude>`,
    and either `<altitudeMode>` or `<gx:altitudeMode>`. The viewing direction of the
    camera is defined by `<heading>`, `<tilt>`, and `<roll>`. `<Camera>` can be a child
    element of any `Feature` or of `<NetworkLinkControl>`. A parent element cannot
    contain both a `<Camera>` and a `<LookAt>` at the same time.

    `<Camera>` provides full six-degrees-of-freedom control over the view, so you can
    position the Camera in space and then rotate it around the X, Y, and Z axes. Most
    importantly, you can tilt the camera view so that you're looking above the horizon
    into the sky.

    `<Camera>` can also contain a :class:`pyLiveKML.objects.TimePrimitive`
    (`<gx:TimeSpan>` or `<gx:TimeStamp>`). Time values in Camera affect historical
    imagery, sunlight, and the display of time-stamped features. For more information,
    read Time with AbstractViews in the Time and Animation chapter of the Developer's Guide.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#camera

    Parameters
    ----------
    viewer_options : ViewerOption | Iterable[ViewerOption] | None, default = None
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None, default = None
        Timestamp or timespan assigned to the object.
    lla : tuple[float, float, float], default = (0, 0, 0)
        The longitude, latitude and altitude (in that order, in decimal degrees) of the
        camera position.
    angles : tuple[float, float, float], default (0, 0, 0)
        The heading, tilt and roll (in that order, in decimal degrees) of the camera
        facing.
    altitude_mode : AltitudeModeEnum | None, default = None

    Attributes
    ----------
    viewer_options : list[ViewerOption]
        Enable or disable one or more Google Earth view modes.
    time_primitive : TimePrimitive | None
        Timestamp or timespan assigned to the object.
    longitude : float
        The longitude of the camera position, in decimal degrees.
    latitude : float
        The latitude of the camera position, in decimal degrees.
    altitude : float
        The altitude of the camera position, in metres, with respect to `altitude_mode`.
    heading : float
        The heading of the camera facing, in decimal degrees.
    tilt : float
        The tilt of the camera facing, in decimal degrees.
    roll : float
        The roll of the camera facing, in decimal degrees.
    altitude_mode : AltitudeModeEnum | None

    """

    _kml_tag = "Camera"
    _kml_fields = AbstractView._kml_fields + (
        _FieldDef("longitude", parser=Angle180),
        _FieldDef("latitude", parser=Angle90),
        _FieldDef("altitude"),
        _FieldDef("heading", parser=Angle360),
        _FieldDef("tilt", parser=AnglePos180),
        _FieldDef("roll", parser=Angle180),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )

    def __init__(
        self,
        viewer_options: Iterable[ViewerOption] | ViewerOption | None = None,
        time_primitive: TimePrimitive | None = None,
        lla: tuple[float, float, float] = (0, 0, 0),
        angles: tuple[float, float, float] = (0, 0, 0),
        altitude_mode: AltitudeModeEnum | None = None,
    ):
        """Camera instance constructor."""
        AbstractView.__init__(self, viewer_options, time_primitive)
        self.longitude, self.latitude, self.altitude = lla
        self.heading, self.tilt, self.roll = angles
        self.altitude_mode = altitude_mode



from pyLiveKML.types.GeoColor import GeoColor
from pyLiveKML.types.GeoCoordinates import GeoCoordinates
from pyLiveKML.types.Vec2 import HotSpot, OverlayXY, ScreenXY, RotationXY, Size
from pyLiveKML.types.ViewerOption import ViewerOption
from pyLiveKML.types.types import (
    AltitudeModeEnum,
    ColorModeEnum,
    DisplayModeEnum,
    FlyToModeEnum,
    GridOriginEnum,
    OverlayShapeEnum,
    PlayModeEnum,
    ViewerOptionEnum,
    ItemIconModeEnum,
    ListItemTypeEnum,
    RefreshModeEnum,
    StyleStateEnum,
    UnitsEnum,
    ViewRefreshModeEnum,
)


__all__ = [
    "AltitudeModeEnum",
    "ColorModeEnum",
    "DisplayModeEnum",
    "FlyToModeEnum",
    "GeoColor",
    "GeoCoordinates",
    "GridOriginEnum",
    "HotSpot",
    "ItemIconModeEnum",
    "ListItemTypeEnum",
    "OverlayShapeEnum",
    "OverlayXY",
    "PlayModeEnum",
    "RefreshModeEnum",
    "RotationXY",
    "ScreenXY",
    "Size",
    "StyleStateEnum",
    "UnitsEnum",
    "ViewRefreshModeEnum",
    "ViewerOption",
    "ViewerOptionEnum",
]



"""AnimatedUpdate module."""

from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.Update import Update, UpdateSequent
from pyLiveKML.objects.Object import _ChildDef, ObjectChild
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class AnimatedUpdate(TourPrimitive):
    """A KML `<gx:AnimatedUpdate>` tag constructor.

    `<AnimatedUpdate>` controls changes during a tour to KML features, using an
    `<Update>`. Changes to KML features will not modify the DOM - that is, any changes
    will be reverted when the tour is over, and will not be saved in the KML at any time.

    `<AnimatedUpdate>` should also contain a `duration` value to specify the length of
    time in seconds over which the update takes place. Integer, float, and color fields
    are smoothly animated from original to new value across the duration; boolean,
    string, and other values that don't lend to interpolation are updated at the end of
    the duration.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxanimatedupdate.

    Parameters
    ----------
    duration : float, default = 0
        Specifies the length of time, in seconds, over which the update takes place.
    delayed_start : float, default = 0
        Specifies the number of seconds to wait (after the inline start position) before
        starting the update.
    target_href : str, default = ""
        The target href for the enclosed :class:`pyLiveKML.objects.Update` instance.
    sequence : UpdateSequent | Iterable[UpdateSequent] | None, default = None
        The sequence of create, change and delete operations to be constructed by the
        enclosed :class:`pyLiveKML.objects.Update` instance.

    Attributes
    ----------
    duration : float, default = 0
        Specifies the length of time, in seconds, over which the update takes place.
    delayed_start : float, default = 0
        Specifies the number of seconds to wait (after the inline start position) before
        starting the update.
    update : Update
        The :class:`pyLiveKML.objects.Update` instance that will generate the animation
        instructions.

    """

    _kml_tag = "gx:AnimatedUpdate"
    _kml_fields = TourPrimitive._kml_fields + (
        _FieldDef("duration", "gx:duration"),
        _FieldDef("delayed_start", "gx:delayedStart"),
    )
    _kml_children = TourPrimitive._kml_children + (_ChildDef("update"),)

    def __init__(
        self,
        duration: float = 0,
        delayed_start: float = 0,
        target_href: str = "",
        sequence: UpdateSequent | Iterable[UpdateSequent] | None = None,
    ) -> None:
        """GxAnimatedUpdate instance constructor."""
        TourPrimitive.__init__(self)
        self.delayed_start = delayed_start
        self.duration = duration
        self.update = Update(target_href, sequence=sequence)



"""MultiTrack module."""

from typing import Iterator, Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Object import _ChildDef, _FieldDef, _ListObject
from pyLiveKML.objects.Track import Track
from pyLiveKML.types.types import AltitudeModeEnum


# TODO: As with MultiGeometry, MultiTrack is in practice a Container, so it must be
# possible to add and delete its contents.


class MultiTrack(_ListObject[Track], Geometry):
    """A KML `<MultiTrack>` tag constructor.

    A multi-track element is used to combine multiple track elements into a single
    conceptual unit. For example, suppose you collect GPS data for a day's bike ride that
    includes several rest stops and a stop for lunch. Because of the interruptions in
    time, one bike ride might appear as four different tracks when the times and
    positions are plotted. Grouping these `Track` elements into one `MultiTrack`
    container causes them to be displayed in Google Earth as sections of a single path.
    When the icon reaches the end of one segment, it moves to the beginning of the next
    segment. The `interpolate` attribute specifies whether to stop at the end of one
    track and jump immediately to the start of the next one, or to interpolate the
    missing values between the two tracks.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#multitrack

    Parameters
    ----------
    tracks : Track | Iterable[Track] | None, default = None
        The `Track` instances to be hosted under this `MultiTrack`.

    Attributes
    ----------
    Nil

    """

    _kml_tag = "gx:MultiTrack"
    _kml_fields = Geometry._kml_fields + (
        _FieldDef("altitude_mode", "gx:altitudeMode"),
        _FieldDef("interpolate", "gx:interpolate"),
    )
    _kml_children = Geometry._kml_children + (_ChildDef("tracks"),)

    def __init__(
        self,
        altitude_mode: AltitudeModeEnum | None = None,
        interpolate: bool = False,
        tracks: Track | Iterable[Track] | None = None,
    ) -> None:
        """MultiTrack instance constructor."""
        _ListObject[Track].__init__(self)
        Geometry.__init__(self)
        self.altitude_mode = altitude_mode
        self.interpolate = interpolate
        self.tracks = tracks

    @property
    def tracks(self) -> Iterator[Track]:
        """The `Tracks` contained by this `MultiTrack` instance.

        Parameters
        ----------
        value : Track | Iterable[Track] | None

        :return: The `Tracks` contained by this `MultiTrack` instance.
        :rtype: Iterator[Track]

        """
        yield from self

    @tracks.setter
    def tracks(self, value: Track | Iterable[Track] | None) -> None:
        self.clear()
        if value is not None:
            if isinstance(value, Track):
                self.append(value)
            else:
                self.extend(value)



"""NetworkLinkControl module."""

from datetime import datetime
from itertools import islice

from lxml import etree  # type: ignore

from pyLiveKML import KML_UPDATE_CONTAINER_LIMIT_DEFAULT
from pyLiveKML.errors import NetworkLinkControlUpdateLimited
from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Update import Update
from pyLiveKML.objects.Container import Container
from pyLiveKML.objects.Folder import Folder
from pyLiveKML.objects.Object import (
    _BaseObject,
    _DependentDef,
    _FieldDef,
    ObjectChild,
    ObjectState,
)


class NetworkLinkControl(_BaseObject):
    """A KML `<NetworkLinkControl>` tag constructor.

    Controls the behavior of files fetched by a `NetworkLink`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#networklinkcontrol

    Parameters
    ----------
    target_href : str
        Used to construct an embedded `Update` instance.
    container : Container | None, default = None
        The root `Container` for synchronization. If `None`, a `Folder` named "root" will
        be created.
    update_limit : int, default = KML_UPDATE_CONTAINER_LIMIT_DEFAULT
        The maximum number of entries per synchronization update.
    min_refresh_period : float | None, default = None
        Specified in seconds, `min_refresh_period` is the minimum allowed time between
        fetches of the file. This parameter allows servers to throttle fetches of a
        particular file and to tailor refresh rates to the expected rate of change to the
        data.
    max_session_length : float | None, default = None
        Specified in seconds, `max_session_length` is the maximum amount of time for
        which the client `NetworkLink` can remain connected. The default value of -1
        indicates not to terminate the session explicitly.
    cookie : str | None, default = None
        Use this element to append a string to the URL query on the next refresh of the
        network link. You can use this data in your script to provide more intelligent
        handling on the server side, including version querying and conditional file
        delivery.
    message : str | None, default = None
        You can deliver a pop-up message, such as usage guidelines for your
        `NetworkLink`. The message appears when the `NetworkLink` is first loaded into
        Google Earth, or when it is changed in the `NetworkLink` control.
    link_name : str | None, default = None
        You can control the name of the `NetworkLink` from the server, so that changes
        made to the name on the client side are overridden by the server.
    link_description : str | None, default = None
        You can control the description of the `NetworkLink` from the server, so that
        changes made to the description on the client side are overridden by the server.
    link_snippet : str | None, default = None
        You can control the snippet for the network link from the server, so that changes
        made to the snippet on the client side are overridden by the server.
    link_snippet_max_lines : int, default = 2
        An integer that specifies the maximum number of snippet lines to display.
    link_expires : datetime | None, default = None
        You can specify a `datetime` at which the link should be refreshed. This
        specification takes effect only if the `refresh_mode` in `Link` has a value of
        `ON_EXPIRE`.
    abstract_view : AbstractView | None, default = None
        A `Camera` or `LookAt` that will set the viewing point when the `NetworkLink`
        loads.

    Attributes
    ----------
    container : Container | None
    update_limit : int
    min_refresh_period : float | None
    max_session_length : float | None
    cookie : str | None
    message : str | None
    link_name : str | None
    link_description : str | None
    link_snippet : str | None
    link_snippet_max_lines : int
    link_expires : datetime | None
    abstract_view : AbstractView | None
    update : Update
        An embedded `Update` instance that is used for synchronization.


    """

    _kml_tag = "NetworkLinkControl"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("min_refresh_period", "minRefreshPeriod"),
        _FieldDef("max_session_length", "maxSessionLength"),
        _FieldDef("cookie"),
        _FieldDef("message"),
        _FieldDef("link_name", "linkName"),
        _FieldDef("link_description", "linkDescription"),
        _FieldDef("link_snippet", "linkSnippet"),
        _FieldDef("link_expires", "linkExpires"),
    )
    _kml_dependents = _BaseObject._kml_dependents + (_DependentDef("update"),)
    _suppress_id = True

    def __init__(
        self,
        target_href: str,
        container: Container | None = None,
        update_limit: int = KML_UPDATE_CONTAINER_LIMIT_DEFAULT,
        min_refresh_period: float | None = None,
        max_session_length: float | None = None,
        cookie: str | None = None,
        message: str | None = None,
        link_name: str | None = None,
        link_description: str | None = None,
        link_snippet: str | None = None,
        link_snippet_max_lines: int = 2,
        link_expires: datetime | None = None,
        abstract_view: AbstractView | None = None,
    ):
        """NetworkLinkControl instance constructor."""
        super().__init__()
        self.container: Container = (
            Folder("Root", is_open=True) if container is None else container
        )
        self.update_limit = update_limit
        self.min_refresh_period = min_refresh_period
        self.max_session_length = max_session_length
        self.cookie = cookie
        self.message = message
        self.link_name = link_name
        self.link_description = link_description
        self.link_snippet = link_snippet
        self.link_snippet_max_lines = link_snippet_max_lines
        self.link_expires = link_expires
        self.abstract_view = abstract_view
        self.update = Update(target_href)

    def construct_sync(
        self,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> etree.Element:
        """Construct a KML synchronization update.

        The real work gets done here.

        Walks the tree under `container`, looking at each object's state, and create,
        update or delete it as necessary.

        Parameters
        ----------
        with_children : bool, default = True
            Whether the `children` of the `NetworkLinkControl`, or any child or
            dependents objects, should be constructed as sub-tags.
        with_dependents : bool, default = True
            Whether the `dependents` of the `NetworkLinkControl`, or any child or
            dependent objects, should be constructed as sub-tags.

        :returns: A `<NetworkLinkControl>` tag, including the synchronization update.
        :rtype: `etree.Element`

        """
        root = etree.Element(self.kml_tag)
        try:
            self._sync_child_objects(self.container)
        except NetworkLinkControlUpdateLimited:
            pass
        super().build_kml(root, with_children, with_dependents)
        return root

    def _sync_child_objects(self, obj: _BaseObject) -> None:
        for d_obj in obj.dependents:
            if d_obj.child.state == ObjectState.CREATING:
                self.update.creates.append(d_obj)
            elif d_obj.child.state == ObjectState.CHANGING:
                self.update.changes.append(d_obj)
            elif d_obj.child.state in (
                ObjectState.DELETE_CHANGED,
                ObjectState.DELETE_CREATED,
            ):
                self.update.deletes.append(d_obj)
            d_obj.child.synchronized()
            if len(self.update) >= self.update_limit:
                raise NetworkLinkControlUpdateLimited
            if isinstance(d_obj.child, _BaseObject):
                self._sync_child_objects(d_obj.child)

        for c_obj in obj.children:
            if c_obj.child.state == ObjectState.CREATING:
                self.update.creates.append(c_obj)
            elif c_obj.child.state == ObjectState.CHANGING:
                self.update.changes.append(c_obj)
            elif c_obj.child.state in (
                ObjectState.DELETE_CHANGED,
                ObjectState.DELETE_CREATED,
            ):
                self.update.deletes.append(c_obj)
            c_obj.child.synchronized()
            if len(self.update) >= self.update_limit:
                raise NetworkLinkControlUpdateLimited
            if isinstance(c_obj.child, _BaseObject):
                self._sync_child_objects(c_obj.child)

        if isinstance(obj, Container):
            limit = self.update_limit - len(self.update)
            deletes = [
                ObjectChild(obj, delobj) for delobj in islice(obj._deleted, limit)
            ]
            self.update.deletes.extend(deletes)
            obj._deleted = obj._deleted[len(deletes) :]

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Build the KML sub-tags for this `NetworkLinkControl` and append it to the provided `etree.Element`.

        Overridden from :class:`pyLiveKML.objects.Object.Object` to perform some
        additional build steps.

        Parameters
        ----------
        root : etree.Element
            The tag into which the sub-tags are to be inserted.
        with_children : bool, default = True
            Whether the `children` of the `NetworkLinkControl`, or any child or
            dependents objects, should be constructed as sub-tags.
        with_dependents : bool, default = True
            Whether the `dependents` of the `NetworkLinkControl`, or any child or
            dependent objects, should be constructed as sub-tags.

        """
        super().build_kml(root, with_children, with_dependents)
        if self.link_snippet is not None:
            attribs = {}
            if self.link_snippet_max_lines is not None:
                attribs["maxLines"] = str(self.link_snippet_max_lines)
            etree.SubElement(root, "Snippet", attribs).text = self.link_snippet



"""Placemark module."""

from typing import Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Geometry import Geometry
from pyLiveKML.objects.Object import ObjectChild
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class Placemark(Feature):
    """A KML `<Placemark>` tag constructor.

    A `Placemark` is a `Feature` with associated `Geometry`. In Google Earth, a
    `Placemark` appears as a list item in the "Places" panel. A `Placemark` with a
    `Point` has an icon associated with it that marks a point on the Earth in the 3D
    viewer.

    Notes
    -----
    In the Google Earth 3D viewer, a `Point` `Placemark` is the only object you can click
    or roll over. Other `Geometry` objects do not have an icon in the 3D viewer. To give
    the user something to click in the 3D viewer, you would need to create a
    `MultiGeometry` object that contains both a `Point` and the other `Geometry` object.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#placemark

    Parameters
    ----------
    geometry : Geometry
        The Geometry to be displayed.
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "Placemark"

    def __init__(
        self,
        geometry: Geometry,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        inline_style: StyleSelector | None = None,
        style_url: str | None = None,
        region: Region | None = None,
    ):
        """Placemark instance constructor."""
        Feature.__init__(
            self,
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=inline_style,
            region=region,
        )
        self.geometry = geometry

    @property
    def dependents(self) -> Iterator[ObjectChild]:
        """A generator over the dependents of the instance.

        Overridden from `_BaseObject` because when using the default behaviour, if
        `geometry` is a `MultiGeometry`, then the `MultiGeometry` itself is never
        yielded.

        :return: A generator over the dependents of the instance.
        :rtype: Iterator[ObjectChild]
        """
        yield ObjectChild(self, self.geometry)

    def activate(self, value: bool, cascade: bool = False) -> None:
        """Activate or deactivate this instance for display in GEP.

        Overridden from `_BaseObject` because when using the default behaviour, the
        superclass method is unaware of `geometry` because it is not flagged as a
        `child`.

        Parameters
        ----------
        value : bool
            `True` for activation, `False` for deactivation.
        cascade : bool, default = False
            `True` if the activation is to be cascaded to all child objects.

        """
        super().activate(value, cascade)
        if cascade:
            self.geometry.activate(value, True)



"""utils module."""

from lxml import etree  # type: ignore


KML_UPDATE_CONTAINER_LIMIT_DEFAULT: int = 100
"""The default value for the container update limit.

The default maximum number of :class:`pyLiveKML.objects.Feature` objects that 
will be included in each synchronization update emitted by a 
:class:`pyLiveKML.objects.NetworkLinkControl` object.
"""


KML_DOCTYPE: str = '<?xml version="1.0" encoding="UTF-8"?>'
"""The XML tag that opens any XML document, including any KML document."""

KML_HEADERS = {"Content-Type": "application/vnd.google-earth.kml+xml"}
"""The headers that should be included when a KML file is tranmitted via HTTP."""

__root_namespace_map = {
    "gx": "http://www.google.com/kml/ext/2.2",
    "kml": "http://www.opengis.net/kml/2.2",
    "atom": "http://www.w3.org/2005/Atom",
}
"""The namespace map that is to applied to all Google Earth KML files."""

__root_attributes = {"xmlns": "http://www.opengis.net/kml/2.2"}
"""The attributes that should be applied to all Google Earth KML files."""


def kml_root_tag() -> etree.Element:
    """Construct the opening <kml> tag, with namespaces, for a KML document.

    :return: The <kml> tag, with namespaces, that encloses the contents of a KML document.
    :rtype: etree.Element
    """
    return etree.Element("kml", nsmap=__root_namespace_map, attrib=__root_attributes)


def with_ns(tag: str) -> str:
    """Transform a Clark-notation XML string into a colon-notation XML string.

    If there is no colon in the text, just return the text. Otherwise, split the text
    into two parts at the first colon. Look up the first part as a key in
    `_root_namespace_map` and return the corresponding value, wrapped in {}, with the
    second part of the original text appended.

    For example, "gx:Track" would return "{http://www.google.com/kml/ext/2.2}Track".
    lxml publishes this as a <gx:Track> tag. Ridiculous double-entry nonsense, but it
    works.

    :return: The transformed input.
    :rtype: str
    """
    parts = tag.split(":", 1)
    return tag if len(parts) < 2 else f"{{{__root_namespace_map[parts[0]]}}}{parts[1]}"



"""Icon module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import RefreshModeEnum, ViewRefreshModeEnum
from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.Link import Link


class Icon(Link):
    """A KML `<Icon>` tag constructor.

    Defines an image associated with an `IconStyle` or `Overlay`. The `href` attribute
    defines the location of the image to be used as the overlay or as the icon for the
    placemark. This location can either be on a local file system or a remote web server.
    The `x`, `y`, `w`, and `h` attributes are used to select one icon from an image that
    contains multiple icons (often referred to as an icon palette).

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#icon

    Parameters
    ----------
    href: str | None, default = None
        An HTTP address or a local file specification used to load an icon.
    refresh_mode: RefreshModeEnum | None, default = None,
        Specifies a time-based refresh mode.
    refresh_interval: float | None, default = None
        Indicates to refresh the file every n seconds.
    view_refresh_mode: ViewRefreshModeEnum, default | None = None
        Specifies how the link is refreshed when the "camera" changes.
    view_refresh_time: float | None, default = None
        After camera movement stops, specifies the number of seconds to wait before
        refreshing the view.
    view_bound_scale: float | None, default = None
        Scales the BBOX parameters before sending them to the server. A value less than 1
        specifies to use less than the full view (screen). A value greater than 1
        specifies to fetch an area that extends beyond the edges of the current view.
    view_format: str | None, default = None
        Specifies the format of the query string that is appended to `href` before the
        file is fetched.
    http_query: str | None, default = None
        Appends information to the query string, based on the parameters specified.
        (Google Earth substitutes the appropriate current value at the time it creates
        the query string.) The following parameters are supported:
        * [clientVersion]
        * [kmlVersion]
        * [clientName]
        * [language]
    x: int, default = 0
        If `href` specifies an icon palette, `x` and `y` identify the offsets, in
        pixels, from the lower-left corner of the icon palette. If no values are
        specified for `x` and `y`, the lower left corner of the icon palette is assumed
        to be the lower-left corner of the icon to use.
    y: int, default = 0
        If `href` specifies an icon palette, `x` and `y` identify the offsets, in
        pixels, from the lower-left corner of the icon palette. If no values are
        specified for `x` and `y`, the lower left corner of the icon palette is assumed
        to be the lower-left corner of the icon to use.
    w: int, default = -1
        If `href` specifies an icon palette, `w` and `h` specify the width and height, in
        pixels, of the icon to use.
    h: int, default = -1
        If `href` specifies an icon palette, `w` and `h` specify the width and height, in
        pixels, of the icon to use.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Icon"
    _kml_fields = Link._kml_fields + (
        _FieldDef("x", "gx:x"),
        _FieldDef("y", "gx:y"),
        _FieldDef("w", "gx:w"),
        _FieldDef("h", "gx:h"),
    )

    def __init__(
        self,
        href: str | None = None,
        refresh_mode: RefreshModeEnum | None = None,
        refresh_interval: float | None = None,  # 4.0
        view_refresh_mode: ViewRefreshModeEnum | None = None,
        view_refresh_time: float | None = None,
        view_bound_scale: float | None = None,
        view_format: str | None = None,
        http_query: str | None = None,
        x: int = 0,
        y: int = 0,
        w: int = -1,
        h: int = -1,
    ):
        """Icon instance constructor."""
        Link.__init__(
            self,
            href=href,
            refresh_mode=refresh_mode,
            refresh_interval=refresh_interval,
            view_refresh_mode=view_refresh_mode,
            view_refresh_time=view_refresh_time,
            view_bound_scale=view_bound_scale,
            view_format=view_format,
            http_query=http_query,
        )
        self.x = x
        self.y = y
        self.w = w
        self.h = h



"""Wait module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class Wait(TourPrimitive):
    """A KML `<gx:Wait>` tag constructor.

    The camera remains still, at the last-defined `AbstractView`, for the number of
    seconds specified before playing the next `TourPrimitive`.

    Notes
    -----
    * A `Wait` does not pause the tour timeline - currently-playing sound files and
    animated updates will continue to play while the camera is waiting.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxwait

    Parameters
    ----------
    duration : float, default = 0
        The time in seconds to wait.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:Wait"
    _kml_fields = TourPrimitive._kml_fields + (_FieldDef("duration"),)

    def __init__(
        self,
        duration: float = 0,
    ) -> None:
        """Wait instance constructor."""
        TourPrimitive.__init__(self)
        self.duration = duration



"""Container module."""

from abc import ABC
from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Object import _ChildDef, _ListObject, ObjectState
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class Container(_ListObject[Feature], Feature, ABC):
    """A KML `<Container>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file. A `Container`
    element holds one or more `Feature` elements, which may themselves be `Container`
    instances, so allows the creation of nested hierarchies.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#container

    Parameters
    ----------
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    features : Feature | Iterable[Feature] | None, default = None
        The `Feature`'s contained by this `Container`.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_children: tuple[_ChildDef, ...] = Feature._kml_children + (
        _ChildDef("features"),
    )

    def __init__(
        self,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        features: Feature | Iterable[Feature] | None = None,
    ):
        """Feature instance constructor."""
        Feature.__init__(
            self,
            name=name,
            is_open=is_open,
            visibility=visibility,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
        )
        _ListObject[Feature].__init__(self)
        ABC.__init__(self)
        self._deleted: list[Feature] = list[Feature]()
        self.features = features
        self._update_limit: int = 0

    @property
    def features(self) -> Iterator[Feature]:
        """Retrieve a generator over the `Features` in this `Container`.

        If the property setter is called, replaces the current list of contained
        `Feature`'s with those provided.

        Parameters
        ----------
        value : Feature | Iterable[Feature] | None
            The new `Feature` elements for the `Container`.

        :returns: A generator over the `Features` in the `Container`.
        :rtype: Iterator[Feature]

        """
        yield from self

    @features.setter
    def features(self, value: Feature | Iterable[Feature] | None) -> None:
        self._deleted.extend(self)
        self.clear()
        if value is not None:
            if isinstance(value, Feature):
                self.append(value)
            else:
                self.extend(value)

    def clear(self) -> None:
        """Remove all of the `Feature`'s enclosed in this `Container`.

        Effectively, moves them to the `_deleted` list.
        """
        self._deleted.extend(self.features)
        super().clear()

    def remove(self, value: Feature) -> None:
        """Remove a single `Feature` from this `Container.

        Effectively, moves it to the `_deleted` list.

        Parameters
        ----------
        value : Feature
            The feature to be removed/moved.

        """
        if value.active:
            self._deleted.append(value)
        super().remove(value)

    def force_idle(self, cascade: bool = False) -> None:
        """Force this instance, and _optionally_ its children, to the `IDLE` state.

        Overridden from :func:`pyLiveKML.objects.Object.Object.force_idle` to enable the
        entire tree of enclosed `Feature` instances, and child objects, that is rooted in
        this `Container` to be forced to the `IDLE` state.

        Parameters
        ----------
        cascade : bool, default = False
            If cascade is `True`, the entire tree under this `Container` is forced
            `IDLE`. If `False`, only the `Container` itself is forced `IDLE`.

        """
        super().force_idle()
        if cascade:
            self.force_features_idle()

    def force_features_idle(self) -> None:
        """Force this instance, and _all_ of its children, to the `IDLE` state.

        Force the entire tree of enclosed `Feature` instances to the `IDLE` state.
        Typically called as a result of the target `Container` being deactivated.
        """
        for f in self:
            if isinstance(f, Container):
                # note the implication from force_idle() that cascade is _always_ true for force_features_idle
                f.force_idle(True)
            elif isinstance(f, Feature):
                f.force_idle()

    def activate(self, value: bool, cascade: bool = False) -> None:
        """Activate the `Container`, i.e. make it appear in GEP.

        Overrides :func:`pyLiveKML.objects.Feature.Feature.activate` to implement
        activate/deactivate cascade to enclosed `Feature` instances, and to ensure that
        if a `Container` is deleted from GEP, its' enclosed `Feature` objects are forced
        `IDLE` to maintain synchronization.

        Notes
        -----
        * Cascade activate upwards, but do not cascade deactivate upwards.

        Parameters
        ----------
        value : bool
            `True` to activate; `False` to deactivate.
        cascade : bool, default = False
            If cascade is `True`, the entire tree under this `Container` is activated. If
            `False`, only the `Container` itself is activated.

        """
        Feature.activate(self, value, cascade)
        if cascade:
            for f in self:
                f.activate(value, True)
        if (
            self._state == ObjectState.DELETE_CREATED
            or self._state == ObjectState.DELETE_CHANGED
        ):
            self._deleted.clear()
            self.force_features_idle()



"""ListStyle module."""

from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.types import ItemIconModeEnum, ListItemTypeEnum, GeoColor
from pyLiveKML.objects.Object import _BaseObject, _ChildDef, _FieldDef, ColorParse
from pyLiveKML.objects.SubStyle import SubStyle


class ItemIcon(_BaseObject):
    """A KML `<ItemIcon>` tag constructor.

    Icon used in the List view that reflects the state of a `Folder` or `Link` fetch.
    Icons associated with the open and closed modes are used for `Folder`s and
    `NetworkLink`s. Icons associated with the error and fetching0, fetching1, and
    fetching2 modes are used for `NetworkLink`s.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-liststyle

    Parameters
    ----------
    icon_state : ItemIconModeEnum | None, default = None
    href : str | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "ItemIcon"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("icon_state", "state"),
        _FieldDef("href"),
    )

    def __init__(
        self,
        icon_state: ItemIconModeEnum | None = None,
        href: str | None = None,
    ):
        """ItemIcon instance constructor."""
        super().__init__()
        self.icon_state = icon_state
        self.href = href


class ListStyle(SubStyle):
    """A KML `<ListStyle>` tag constructor.

    Specifies how a `Feature` is displayed in the list view. The list view is a hierarchy
    of containers and children; in Google Earth, this is the "Places" panel.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#liststyle

    Parameters
    ----------
    list_item_type: ListItemTypeEnum | None, default = None
        Specifies how a Feature is displayed in the list view.
    bg_color: GeoColor | int | None, default = None
        Background color for the `Feature`'s `snippet`.
    icons: ItemIcon | Iterable[ItemIcon] | None, default = None
        Mappings between the state of the list item, and the icon to be displayed in the
        "Places" panel.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "ListStyle"
    _kml_fields = SubStyle._kml_fields + (
        _FieldDef("list_item_type", "listItemType"),
        _FieldDef("bg_color", "bgColor", ColorParse),
    )
    _kml_children = SubStyle._kml_children + (_ChildDef("icons"),)

    def __init__(
        self,
        list_item_type: ListItemTypeEnum | None = None,
        bg_color: GeoColor | int | None = None,
        icons: ItemIcon | Iterable[ItemIcon] | None = None,
    ):
        """ListStyle instance constructor."""
        SubStyle.__init__(self)
        self.list_item_type = list_item_type
        self.bg_color = bg_color
        self._icons = list[ItemIcon]()
        self.icons = icons

    @property
    def icons(self) -> Iterable[ItemIcon]:
        """Retrieve a generator over the `icons` of this `ListStyle`.

        If the property setter is called, replaces the current list of icons with
        those provided.

        Parameters
        ----------
        value : ItemIcon | Iterable[ItemIcon] | None
            The new icons for the `ListStyle`.

        :returns: A generator over the `icons` of the `ListStyle`.
        :rtype: Iterator[ItemIcon]

        """
        yield from self._icons

    @icons.setter
    def icons(self, value: ItemIcon | Iterable[ItemIcon] | None) -> None:
        self._icons.clear()
        if value is not None:
            if isinstance(value, ItemIcon):
                self._icons.append(value)
            else:
                self._icons.extend(value)



"""Link module."""

from lxml import etree  # type: ignore

from pyLiveKML.types import RefreshModeEnum, ViewRefreshModeEnum
from pyLiveKML.objects.Object import _FieldDef, Object


class Link(Object):
    """A KML `<Link>` tag constructor.

    Specifies the location of any of the following:
    * KML files fetched by network links.
    * Image files used in any `Overlay`.
    * Model files used in `Model`.

    The file is conditionally loaded and refreshed, depending on the refresh parameters
    supplied here. Two different sets of refresh parameters can be specified: one set is
    based on time (`refresh_mode` and `refresh_interval`) and one is based on the current
    "camera" view (`view_refresh_mode` and `view_refresh_time`). In addition, `Link`
    specifies whether to scale the bounding box parameters that are sent to the server
    (`view_bound_scale`) and provides a set of optional viewing parameters that can be
    sent to the server (`view_format`) as well as a set of optional parameters containing
    version and language information.

    When a file is fetched, the URL that is sent to the server is composed of three
    pieces of information:

    * The href that specifies the file to load.
    * An arbitrary format string that is created from:
        * parameters that you specify in the `view_format` attribute, or
        * bounding box parameters (this is the default and is used if no `view_format`
        attribute is set)
    * A second format string that is specified in the `http_query` attribute.

    If the file specified in `href` is a local file, the `view_format` and `http_query`
    attributes are not used.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#icon

    Parameters
    ----------
    href: str | None, default = None
        An HTTP address or a local file specification used to load an icon.
    refresh_mode: RefreshModeEnum | None, default = None,
        Specifies a time-based refresh mode.
    refresh_interval: float | None, default = None
        Indicates to refresh the file every n seconds.
    view_refresh_mode: ViewRefreshModeEnum, default | None = None
        Specifies how the link is refreshed when the "camera" changes.
    view_refresh_time: float | None, default = None
        After camera movement stops, specifies the number of seconds to wait before
        refreshing the view.
    view_bound_scale: float | None, default = None
        Scales the BBOX parameters before sending them to the server. A value less than 1
        specifies to use less than the full view (screen). A value greater than 1
        specifies to fetch an area that extends beyond the edges of the current view.
    view_format: str | None, default = None
        Specifies the format of the query string that is appended to `href` before the
        file is fetched.
    http_query: str | None, default = None
        Appends information to the query string, based on the parameters specified.
        (Google Earth substitutes the appropriate current value at the time it creates
        the query string.) The following parameters are supported:
        * [clientVersion]
        * [kmlVersion]
        * [clientName]
        * [language]

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Link"
    _kml_fields = Object._kml_fields + (
        _FieldDef("href"),
        _FieldDef("refresh_mode", "refreshMode"),
        _FieldDef("refresh_interval", "refreshInterval"),
        _FieldDef("view_refresh_mode", "viewRefreshMode"),
        _FieldDef("view_refresh_time", "viewRefreshTime"),
        _FieldDef("view_bound_scale", "viewBoundScale"),
        _FieldDef("view_format", "viewFormat"),
        _FieldDef("http_query", "httpQuery"),
    )

    def __init__(
        self,
        href: str | None = None,
        refresh_mode: RefreshModeEnum | None = None,
        refresh_interval: float | None = None,
        view_refresh_mode: ViewRefreshModeEnum | None = None,
        view_refresh_time: float | None = None,
        view_bound_scale: float | None = None,
        view_format: str | None = None,
        http_query: str | None = None,
    ):
        """Link instance constructor."""
        Object.__init__(self)
        self.href = href
        self.refresh_mode = refresh_mode
        self.refresh_interval = refresh_interval
        self.view_refresh_mode = view_refresh_mode
        self.view_refresh_time = view_refresh_time
        self.view_bound_scale = view_bound_scale
        self.view_format = view_format
        self.http_query = http_query



"""KML module."""

import enum


class AltitudeModeEnum(enum.Enum):
    """Enumeration of options for KML `<gx:altitudeMode>` tags.

    Generally used in e.g. objects that derive from :class:`pyLiveKML.objects.Geometry`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields

    Notes
    -----
    * For simplicity, all `<altitudeMode>` tags in pyLiveKML have been replaced with
    `<gx:altitudeMode>` tags to ease implementation of the seafloor-based modes.

    """

    CLAMP_TO_SEAFLOOR = "clampToSeaFloor"
    RELATIVE_TO_SEAFLOOR = "relativeToSeaFloor"
    CLAMP_TO_GROUND = "clampToGround"
    RELATIVE_TO_GROUND = "relativeToGround"
    ABSOLUTE = "absolute"


class ColorModeEnum(enum.Enum):
    """Enumeration of options for KML `<colorMode>` tags.

    Specifically for objects that derive from :class:`pyLiveKML.objects.ColorStyle`.

    A value of `RANDOM` applies a random linear scale to the base `<color>` as follows.

    * To achieve a truly random selection of colors, specify a base `<color>` of solid
    white (0xffffffff).
    * If you specify a single color component (for example, a value of 0xff0000ff for
    solid red), random color values for that one component (red) will be selected. In
    this case, the values would range from 0x00 (black) to 0xff (full red).
    * If you specify values for two or for all three color components, a random linear
    scale is applied to each color component, with results ranging from black to the
    maximum values specified for each component.
    * The opacity of a color comes from the alpha component of `<color>` and is never
    randomized.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    NORMAL = "normal"
    RANDOM = "random"


class DisplayModeEnum(str, enum.Enum):
    """Enumeration of options for KML `<displayMode>` tags.

    Specifically for :class:`pyLiveKML.objects.BalloonStyle` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    DEFAULT = "default"
    HIDE = "hide"


class FlyToModeEnum(enum.Enum):
    """Enumeration of options for KML `<gx:FlyTo>` tags.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxflyto.

    """

    BOUNCE = "bounce"
    SMOOTH = "smooth"


class PlayModeEnum(enum.Enum):
    """Enumeration of options for KML `<gx:playMode>` tags.

    Used only by :class:`pyLiveKML.objects.TourControl`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gxtourcontrol.

    """

    PAUSE = "pause"


class ViewerOptionEnum(enum.Enum):
    """Enumeration of options for KML `<gx:option>` tags.

    Used only by subclasses of :class:`pyLiveKML.objects.AbstractView`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#abstractview.

    """

    STREETVIEW = "streetview"
    HISTORICAL_IMAGERY = "historicalimagery"
    SUNLIGHT = "sunlight"
    GROUND_NAVIGATION = "groundnavigation"


class ItemIconModeEnum(str, enum.Enum):
    """Enumeration of options for KML `<ItemIcon><state>` tags.

    Specifically for :class:`pyLiveKML.objects.ListStyle` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    OPEN = "open"
    CLOSED = "closed"
    ERROR = "error"
    OPEN_ERROR = "open error"
    CLOSED_ERROR = "closed error"
    OPEN_FETCHING0 = "open fetching0"
    OPEN_FETCHING1 = "open fetching1"
    OPEN_FETCHING2 = "open fetching2"
    CLOSED_FETCHING0 = "closed fetching0"
    CLOSED_FETCHING1 = "closed fetching1"
    CLOSED_FETCHING2 = "closed fetching2"


class ListItemTypeEnum(str, enum.Enum):
    """Enumeration of options for KML `<listItemType>` tags.

    Specifically for :class:`pyLiveKML.objects.ListStyle` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    CHECK = "check"
    CHECK_OFF_ONLY = "checkOffOnly"
    CHECK_HIDE_CHILDREN = "checkHideChildren"
    RADIO_FOLDER = "radioFolder"


class RefreshModeEnum(str, enum.Enum):
    """Enumeration of options for KML `<refreshMode>` tags.

    Specifically for :class:`pyLiveKML.objects.Link` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    ON_CHANGE = "onChange"
    ON_INTERVAL = "onInterval"
    ON_EXPIRE = "onExpire"


class StyleStateEnum(str, enum.Enum):
    """Enumeration of options for KML `<Pair><key>` tags.

    Specifically for :class:`pyLiveKML.objects.StyleMap` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    NORMAL = "normal"
    HIGHLIGHT = "highlight"


class UnitsEnum(str, enum.Enum):
    """Enumeration of options for KML `<unitsEnum>` tags.

    Specifically for subclasses of :class:`pyLiveKML.objects.Vec2` instances.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    FRACTION = "fraction"
    PIXELS = "pixels"
    INSET_PIXELS = "insetPixels"


class ViewRefreshModeEnum(enum.Enum):
    """Enumeration of options for KML `<viewRefreshMode>` tags.

    Specifically for :class:`pyLiveKML.objects.Link` objects.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#kml-fields.

    """

    NEVER = "never"
    ON_STOP = "onStop"
    ON_REQUEST = "onRequest"
    ON_REGION = "onRegion"


class OverlayShapeEnum(enum.Enum):
    """Enumeration of view shape options for KML `<Overlay>` subclass tags.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#photooverlay.

    """

    RECTANGLE = "rectangle"
    CYLINDER = "cylinder"
    SPHERE = "sphere"


class GridOriginEnum(enum.Enum):
    """Enumeration of grid origin options for KML `<Overlay>` subclass tags.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#photooverlay.

    """

    LOWER_LEFT = "lowerLeft"
    UPPER_LEFT = "upperLeft"



"""Overlay module."""

from abc import ABC
from typing import Iterable, cast

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _FieldDef,
    ColorParse,
    NoDump,
)
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import GeoColor


class _Overlay_Icon(_BaseObject):
    """A minimalist Icon class, used only within `Overlay`."""

    _kml_tag = "Icon"
    _kml_fields = _BaseObject._kml_fields + (_FieldDef("href"),)

    def __init__(self, href: str):
        """_Overlay_Icon instance constructor."""
        super().__init__()
        self.href = href


class Overlay(Feature, ABC):
    """A KML 'Overlay' tag constructor.

    This is an abstract element and cannot be used directly in a KML file. `Overlay` is
    the base type for image overlays drawn on the planet surface or on the screen.
    `icon` specifies the image to use and can be configured to reload images based on a
    timer or by camera changes. This class also includes specifications for stacking
    order of multiple overlays and for adding color and transparency values to the base
    image.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#overlay

    Parameters
    ----------
    icon : str | None, default = None
        Defines the location of the image to be used as the `Overlay`. This location can
        be either on a local file system or on a web server. If this attribute is `None`,
        a rectangle is drawn using the color and size defined by the ground or screen
        overlay.
    color : GeoColor | int | None, default = None
        The image or rectangle color and transparency.
    draw_order : int | None, default = None
        Defines the stacking order for the images in overlapping `Overlay`s. `Overlay`s
        with higher `draw_order` values are drawn on top of `Overlay`s with lower
        `draw_order` values.
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = ""
    _kml_fields: tuple[_FieldDef, ...] = Feature._kml_fields + (
        _FieldDef("icon", dumper=NoDump),
        _FieldDef("color", parser=ColorParse),
        _FieldDef("draw_order", "drawOrder"),
    )
    _kml_children = Feature._kml_children + (_ChildDef("icon"),)

    def __init__(
        self,
        icon: str | None = None,
        color: GeoColor | int | None = None,
        draw_order: int | None = None,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
    ):
        """Overlay instance constructor."""
        Feature.__init__(
            self,
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
        )
        ABC.__init__(self)
        self._icon = _Overlay_Icon(icon) if icon else None
        self.draw_order = draw_order
        self.color = cast(GeoColor | None, color)

    @property
    def icon(self) -> str | None:
        """The location of the icon image file.

        Parameters
        ----------
        value : str | None

        :return: The (optional) location of the icon image file.
        :rtype: str | None

        """
        return self._icon.href if self._icon else None

    @icon.setter
    def icon(self, value: str | None) -> None:
        if value is None:
            self._icon = None
        elif self._icon is None:
            self._icon = _Overlay_Icon(value)
        else:
            self._icon.href = value
